// Generated by CoffeeScript 2.3.2
(function() {
  var GedaGenerator, fs, log, mkdirp, sprintf;

  fs = require('fs');

  mkdirp = require('mkdirp');

  sprintf = require('sprintf-js').sprintf;

  log = require('./qeda-log');

  
  // Generator of gEDA gschem symbols and gEDA pcb footprints
  // gEDA gschem symbol file format is documented at http://wiki.geda-project.org/geda:file_format_spec
  // gEDA pcb footprint file format is documented at http://pcb.geda-project.org/pcb-cvs/pcb.html#File-Formats

  GedaGenerator = class GedaGenerator {
    
    // Constructor

    constructor(library) {
      this.library = library;
      this.f = `%.${this.library.pattern.decimals}f`;
    }

    
    // Generate symbol files

    generate(name1) {
      var dir, element, fd, j, k, len, len1, ref, ref1, results;
      this.name = name1;
      // output directory for symbols
      dir = `./geda/symbols/${this.name}/`;
      mkdirp.sync(`${dir}`);
      ref = this.library.elements;
      // Symbols
      for (j = 0, len = ref.length; j < len; j++) {
        element = ref[j];
        if ((element.symbols == null) || 0 === element.symbols.length) {
          continue;
        }
        log.start(`gEDA gschem symbol '${element.name}'`);
        this._generateSymbol(dir, element);
        log.ok();
      }
      // output directory for footprints
      dir = `./geda/footprints/${this.name}/`;
      mkdirp.sync(`${dir}`);
      ref1 = this.library.elements;
      // Footprints
      results = [];
      for (k = 0, len1 = ref1.length; k < len1; k++) {
        element = ref1[k];
        if (element.pattern == null) {
          next;
        }
        log.start(`gEDA pcb footprint '${element.pattern.name}.fp'`);
        fd = fs.openSync(`${dir}/${element.pattern.name}.fp`, 'w');
        this._generatePattern(fd, element.pattern);
        fs.closeSync(fd);
        results.push(log.ok());
      }
      return results;
    }

    
    // Write gEDA gschem symbol(s)

    _generateSymbol(dir, element) {
      var alignment, angle, attribute, color, combined, fd, i, j, k, l, len, len1, len2, min_x, min_y, name, part_name, pintype, ref, ref1, ref2, results, shape, symbol, value, x1, x2, xt, y1, y2, yt;
      // qeda defines one symbol per element "part"
      // the equivalent gEDA gschem mechanism is called "slot"
      // pins can be different per slot, but not graphical elements
      // but qeda defines one rectangle per part, and it's not easy to find which pins are re-used
      // thus we will generate one symbol per part
      // see gEDA gschem convention for more information: http://wiki.geda-project.org/geda:gschem_symbol_creation
      combined = fs.openSync(`${dir}/${element.name}.sym`, 'w');
      fs.writeSync(combined, "v 20150930 2\n"); // use last know file format date
      ref = element.symbols;
      results = [];
      for (j = 0, len = ref.length; j < len; j++) {
        symbol = ref[j];
        symbol.invertVertical(); // Positive vertical axis is pointing up
        symbol.alignToGrid(10);
        min_x = 0;
        min_y = 0;
        ref1 = symbol.shapes;
        for (k = 0, len1 = ref1.length; k < len1; k++) {
          shape = ref1[k];
          if (!((shape.x != null) && (shape.y != null))) {
            continue;
          }
          if (shape.x < min_x) {
            min_x = shape.x;
          }
          if (shape.y < min_y) {
            min_y = shape.y;
          }
        }
        min_x = symbol.alignToGrid(min_x, 'floor');
        min_y = symbol.alignToGrid(min_y, 'floor');
        symbol.resize(100, true, -min_x, -min_y); // Resize to grid 100 (mil), and translate to 0, 0 (enforce at least this gEDA gschem convention)
        part_name = "";
        if (element.symbols.length > 1) {
          part_name = `_part-${element.symbols.indexOf(symbol) + 1}-${element.symbols.length}`;
          if (symbol.name) {
            part_name += `-${symbol.name}`;
          }
        }
        fd = fs.openSync(`${dir}/${element.name}${part_name}.sym`, 'w');
        fs.writeSync(fd, "v 20150930 2\n");
        ref2 = ['datasheet', 'keywords', 'aliases'];
        for (l = 0, len2 = ref2.length; l < len2; l++) {
          attribute = ref2[l];
          if (element[attribute] == null) {
            continue;
          }
          if (element[attribute] instanceof Array) {
            if (element[attribute].length > 0) {
              fs.writeSync(fd, "T 0 0 5 8 0 0 0 0 1\n");
              fs.writeSync(fd, `${attribute}=${element[attribute].join(' ')}\n`);
            }
          } else {
            fs.writeSync(fd, "T 0 0 5 8 0 0 0 0 1\n");
            fs.writeSync(fd, `${attribute}=${element[attribute]}\n`);
          }
        }
        if ((element.pattern != null) && (element.pattern.name != null)) {
          fs.writeSync(fd, "T 0 0 5 8 0 0 0 0 1\n");
          fs.writeSync(fd, `footprint=${element.pattern.name}.fp\n`);
        }
        results.push((function() {
          var len3, m, ref3, ref4, ref5, results1;
          ref3 = symbol.shapes;
          results1 = [];
          for (m = 0, len3 = ref3.length; m < len3; m++) {
            shape = ref3[m];
            // ignore all style attributes (lineWidth, fill, fontSize, ...) to keep it simple, and they are not commonly used
            switch (shape.kind) {
              case 'rectangle':
                results1.push(fs.writeSync(fd, `B ${shape.x1} ${shape.y1} ${shape.x2 - shape.x1} ${shape.y2 - shape.y1} 3 0 1 0 -1 -1 0 -1 -1 -1 -1 -1\n`));
                break;
              case 'line':
                results1.push(fs.writeSync(fd, `L ${shape.x1} ${shape.y1} ${shape.x2} ${shape.y2} 3 0 1 0 -1 -1\n`));
                break;
              case 'circle':
                results1.push(fs.writeSync(fd, `V ${shape.x} ${shape.y} ${shape.radius} 3 0 0 0 -1 -1 0 -1 -1 -1 -1 -1\n`));
                break;
              case 'arc':
                results1.push(fs.writeSync(fd, `A ${shape.x} ${shape.y} ${shape.radius} ${shape.start} ${shape.end - shape.start} 3 0 0 0 -1 -1\n`));
                break;
              case 'pin':
                x1 = shape.x;
                y1 = shape.y;
                x2 = x1;
                y2 = y1;
                switch (shape.orientation) {
                  case 'right':
                    x2 += shape.length;
                    break;
                  case 'left':
                    x2 -= shape.length;
                    break;
                  case 'up':
                    y2 += shape.length;
                    break;
                  case 'down':
                    y2 -= shape.length;
                }
                fs.writeSync(fd, `P ${x1} ${y1} ${x2} ${y2} 1 0 0\n`);
                // write pin attributes
                fs.writeSync(fd, "{\n");
                // pin label and type
                xt = x2;
                yt = y2;
                angle = 0;
                alignment = 0;
                switch (shape.orientation) {
                  case 'right':
                    xt += shape.space;
                    break;
                  case 'left':
                    xt -= shape.space;
                    alignment = 6;
                    break;
                  case 'up':
                    yt += shape.space;
                    angle = 90;
                    break;
                  case 'down':
                    yt -= shape.space;
                    angle = 90;
                    alignment = 6;
                }
                name = shape.name;
                if (name.includes('~') || shape.inverted) {
                  name = name.replace(/^~/, '').replace(/~$/, '');
                  name = `\\_${name}\\_`;
                }
                fs.writeSync(fd, `T ${xt} ${yt} 9 8 ${(element.schematic.showPinNames ? 1 : 0)} 1 ${angle} ${alignment} 1\n`);
                fs.writeSync(fd, `pinlabel=${name}\n`);
                pintype = false;
                if (shape.bidir) {
                  pintype = "io";
                } else if (shape.ground || shape.power) {
                  pintype = "pwr";
                } else if (shape.in) {
                  pintype = "in";
                } else if (shape.out) {
                  pintype = "out";
                } else if (shape.passive) {
                  pintype = "pas";
                } else if (shape.z) {
                  pintype = "tri";
                }
                if (pintype != null) {
                  fs.writeSync(fd, `T ${xt} ${yt} 5 8 0 0 ${angle} ${alignment + 2} 1\n`);
                  fs.writeSync(fd, `pintype=${pintype}\n`);
                }
                // pin number and sequence
                xt = x2;
                yt = y2;
                angle = 0;
                alignment = 0;
                switch (shape.orientation) {
                  case 'right':
                    xt -= shape.space;
                    alignment = 6;
                    break;
                  case 'left':
                    xt += shape.space;
                    break;
                  case 'up':
                    yt -= shape.space;
                    angle = 90;
                    alignment = 6;
                    break;
                  case 'down':
                    yt += shape.space;
                    angle = 90;
                }
                fs.writeSync(fd, `T ${xt} ${yt} 5 8 ${(element.schematic.showPinNumbers ? 1 : 0)} 1 ${angle} ${alignment} 1\n`);
                fs.writeSync(fd, `pinnumber=${shape.number}\n`);
                fs.writeSync(fd, `T ${xt} ${yt} 5 8 0 0 ${angle} ${alignment + 2} 1\n`);
                fs.writeSync(fd, `pinseq=${shape.number}\n`);
                // end pin attributes
                results1.push(fs.writeSync(fd, "}\n"));
                break;
              case 'poly':
                fs.writeSync(fd, `H 3 0 0 0 -1 -1 0 -1 -1 -1 -1 -1 ${shape.points.length / 2 + 1}\n`);
                ref5 = (function() {
                  var results2 = [];
                  for (var n = 0, ref4 = shape.points.length - 1; 0 <= ref4 ? n <= ref4 : n >= ref4; 0 <= ref4 ? n++ : n--){ results2.push(n); }
                  return results2;
                }).apply(this);
                for (i of ref5) {
                  if (i % 2 === 0) {
                    fs.writeSync(fd, `${(i === 0 ? 'M' : 'L')} ${shape.points[i]} ${shape.points[i + 1]}\n`);
                  }
                }
                results1.push(fs.writeSync(fd, "z\n"));
                break;
              case 'text':
                alignment = (function() {
                  switch (shape.valign) {
                    case 'bottom':
                      return 0;
                    case 'center':
                      return 1;
                    case 'top':
                      return 2;
                    default:
                      return 0;
                  }
                })();
                alignment += (function() {
                  switch (shape.halign) {
                    case 'left':
                      return 0;
                    case 'center':
                      return 3;
                    case 'right':
                      return 6;
                    default:
                      return 0;
                  }
                })();
                fs.writeSync(fd, `T ${shape.x} ${shape.y} 9 10 1 0 0 ${alignment} 1\n`);
                results1.push(fs.writeSync(fd, `${shape.text}\n`));
                break;
              case 'attribute':
                alignment = (function() {
                  switch (shape.valign) {
                    case 'bottom':
                      return 0;
                    case 'center':
                      return 1;
                    case 'top':
                      return 2;
                    default:
                      return 0;
                  }
                })();
                alignment += (function() {
                  switch (shape.halign) {
                    case 'left':
                      return 0;
                    case 'center':
                      return 3;
                    case 'right':
                      return 6;
                    default:
                      return 0;
                  }
                })();
                color = shape.name === 'name' ? 9 : 5;
                name = (function() {
                  switch (shape.name) {
                    case 'refDes':
                      return 'refdes';
                    case 'name':
                      return 'device';
                    default:
                      return shape.name.toLowerCase;
                  }
                })();
                if (element[shape.name]) {
                  fs.writeSync(fd, `T ${shape.x} ${shape.y} ${color} 10 1 1 0 ${alignment} 1\n`);
                  value = element[shape.name];
                  if (name === 'refdes') {
                    value += "?";
                  }
                  results1.push(fs.writeSync(fd, `${name}=${value}\n`));
                } else {
                  results1.push(void 0);
                }
                break;
              default:
                results1.push(void 0);
            }
          }
          return results1;
        })());
      }
      return results;
    }

    
    // Write gEDA pcb pattern file

    _generatePattern(fd, pattern) {
      var base, flags, j, len, mask, now, ref, shape, timestamp, width, x1, x2, y1, y2;
      now = new Date;
      timestamp = sprintf("%d-%02d-%02d %02d:%02d:%02d", now.getYear() + 1900, now.getMonth() + 1, now.getDate(), now.getHours(), now.getMinutes(), now.getSeconds());
      fs.writeSync(fd, `# footprint generated using qeda on ${timestamp}\n`);
      if ((base = pattern.attributes.refDes).text == null) {
        base.text = "REF**";
      }
      fs.writeSync(fd, 'Element ["" "' + pattern.name + '" "' + pattern.attributes.refDes.text + '" "" 0 0 0 0 0 100 ""]' + "\n");
      fs.writeSync(fd, '(' + "\n");
      ref = pattern.shapes;
      for (j = 0, len = ref.length; j < len; j++) {
        shape = ref[j];
        switch (shape.kind) {
          case 'attribute':
            if (!shape.text) {
              continue;
            }
            fs.writeSync(fd, sprintf("  Attribute (\"%s\" \"%s\")\n", shape.name, shape.text));
            break;
          case 'circle':
            if (!shape.layer.includes("topSilkscreen")) {
              continue;
            }
            fs.writeSync(fd, sprintf(`  ElementArc [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm 0 360 ${this.f}mm]\n`, shape.x, shape.y, shape.radius, shape.radius, shape.lineWidth));
            break;
          case 'line':
            if (!shape.layer.includes("topSilkscreen")) {
              continue;
            }
            fs.writeSync(fd, sprintf(`  ElementLine [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm]\n`, shape.x1, shape.y1, shape.x2, shape.y2, shape.lineWidth));
            break;
          case 'pad':
            if (shape.clearance == null) {
              shape.clearance = pattern.settings.clearance.padToPad;
            }
            if (shape.mask == null) {
              shape.mask = pattern.settings.clearance.padToMask;
            }
            if (shape.type === 'through-hole' || shape.type === 'mounting-hole') {
              if (shape.width === shape.height) {
                flags = [];
                if (shape.hole >= shape.width) {
                  flags.push('hole');
                }
                if (shape.shape === 'rectangle') {
                  flags.push('square');
                }
                fs.writeSync(fd, sprintf(`  Pin [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm "" "%s" "%s"]\n`, shape.x, shape.y, shape.width, shape.clearance, shape.width + 2 * shape.mask, shape.hole, shape.name, flags.join(',')));
              } else {
                width = shape.width < shape.height ? shape.width : shape.height;
                fs.writeSync(fd, sprintf(`  Pin [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm "" "%s" ``]\n`, shape.x, shape.y, width, shape.clearance, width + 2 * shape.mask, shape.hole, shape.name));
                // let the next section handle drawing the non-square pad
                shape.type = 'smd';
              }
            }
            if (shape.type === 'smd') {
              if (pattern.settings.smoothPadCorners && pattern.settings.maximum.cornerRadius > 0.0) {
                shape.shape = 'rounded';
              }
              width = 0;
              x1 = shape.x;
              x2 = shape.x;
              y1 = shape.y;
              y2 = shape.y;
              if (shape.width < shape.height) {
                width = shape.width;
                y1 = shape.y - shape.height / 2.0 + shape.width / 2.0;
                y2 = shape.y + shape.height / 2.0 - shape.width / 2.0;
              } else {
                width = shape.height;
                x1 = shape.x - shape.width / 2.0 + shape.height / 2.0;
                x2 = shape.x + shape.width / 2.0 - shape.height / 2.0;
              }
              mask = width + 2 * shape.mask;
              if (shape.layer.includes("topCopper")) {
                flags = [];
                if (shape.shape === 'rectangle') {
                  flags.push('square');
                }
                if (!shape.layer.includes("topMask")) {
                  mask = 0;
                }
                if (!shape.layer.includes("topPaste")) {
                  flags.push('nopaste');
                }
                fs.writeSync(fd, sprintf(`  Pad [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm "" "%s" "%s"]\n`, x1, y1, x2, y2, width, shape.clearance, mask, shape.name, flags.join(',')));
              }
              if (shape.layer.includes("bottomCopper")) {
                flags = ['solder'];
                if (shape.shape === 'rectangle') {
                  flags.push('square');
                }
                if (!shape.layer.includes("bottomMask")) {
                  mask = 0;
                }
                if (!shape.layer.includes("bottomPaste")) {
                  flags.push('nopaste');
                }
                fs.writeSync(fd, sprintf(`  Pad [${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm ${this.f}mm "" "%s" "%s"]\n`, x1, y1, x2, y2, width, shape.clearance, mask, shape.name, flags.join(',')));
              }
            }
        }
      }
      return fs.writeSync(fd, ")\n"); // end of Element
    }

  };

  module.exports = GedaGenerator;

}).call(this);
