// Generated by CoffeeScript 2.3.2
(function() {
  var QedaElement, QedaPattern, QedaSymbol, fs, log, yaml;

  fs = require('fs');

  yaml = require('js-yaml');

  QedaSymbol = require('./qeda-symbol');

  QedaPattern = require('./qeda-pattern');

  log = require('./qeda-log');

  
  // Class for electronic component

  QedaElement = class QedaElement {
    
    // Constructor

    constructor(library, description) {
      var alias, group, groups, i, j, key, l, last, len, len1, len2, len3, len4, m, name, newAliases, o, option, options, p, part, pin, pins, q, r, ref, ref1, ref2, ref3, ref4, ref5, ref6, ref7, ref8, s, suffixes, value;
      this.library = library;
      this.mergeObjects(this, description);
      if (this.suffix != null) {
        if (this.aliases == null) {
          this.aliases = [];
        }
        newAliases = [];
        suffixes = this.suffix.replace(/\s+/g, '').split(',');
        newAliases = newAliases.concat(suffixes.map((v) => {
          return this.name + v;
        }));
        ref = this.aliases;
        for (l = 0, len = ref.length; l < len; l++) {
          alias = ref[l];
          newAliases = newAliases.concat(suffixes.map((v) => {
            return alias + v;
          }));
        }
        newAliases = newAliases.filter((v) => {
          return this.aliases.indexOf(v) === -1;
        });
        this.aliases = this.aliases.concat(newAliases);
      }
      // Find longest alias
      this.longestAlias = this.name;
      if (this.alias != null) {
        ref1 = this.alias;
        for (m = 0, len1 = ref1.length; m < len1; m++) {
          alias = ref1[m];
          if (alias.length > this.longestAlias.length) {
            this.longestAlias = alias;
          }
        }
      }
      this.refDes = 'REF'; // Should be overriden in element handler
      this.symbols = []; // Array of symbols (one for single part or several for multi-part)
      this.pins = []; // Array of pin objects
      this.pinGroups = []; // Array of pin groups
      
      // Grid-array row letters
      this.gridLetters = ['', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'L', 'M', 'N', 'P', 'R', 'T', 'U', 'V', 'W', 'Y', 'Z'];
      // Make pairs
      last = this.gridLetters.length - 1;
      for (i = o = 1, ref2 = last; (1 <= ref2 ? o <= ref2 : o >= ref2); i = 1 <= ref2 ? ++o : --o) {
        for (j = p = ref3 = i, ref4 = last; (ref3 <= ref4 ? p <= ref4 : p >= ref4); j = ref3 <= ref4 ? ++p : --p) {
          this.gridLetters.push(this.gridLetters[i] + this.gridLetters[j]);
        }
      }
      // Process joint names
      this.delimiter = {};
      ref5 = this.joint;
      for (key in ref5) {
        value = ref5[key];
        groups = this.parseMultiple(key);
        for (q = 0, len2 = groups.length; q < len2; q++) {
          group = groups[q];
          this.delimiter[group] = value;
        }
      }
      // Create pin objects and groups
      if (typeof this.pinout === 'string') { // Unnamed pins
        this.schematic.simpified = true;
        pins = this.parseMultiple(this.pinout);
        for (r = 0, len3 = pins.length; r < len3; r++) {
          pin = pins[r];
          this._addPins(pin, pin);
        }
      } else {
        this._addPins('', this.pinout);
      }
      ref6 = this.groups;
      // Forming groups
      for (key in ref6) {
        value = ref6[key];
        this.pinGroups[key] = this._concatenateGroups(value);
      }
      this._parseProperties();
      if (((ref7 = this.schematic) != null ? ref7.options : void 0) != null) {
        options = this.schematic.options.replace(/\s+/g, '').toLowerCase().split(',');
        for (s = 0, len4 = options.length; s < len4; s++) {
          option = options[s];
          this.schematic[option] = true;
        }
      }
      if (this.parts != null) {
        ref8 = this.parts;
        for (name in ref8) {
          part = ref8[name];
          this.symbols.push(new QedaSymbol(this, name, this.parseMultiple(part))); // Single-part element
        }
      } else {
        this.symbols.push(new QedaSymbol(this)); // Multi-part element
      }
      // Create pattern
      if (this.housing != null) {
        this.pattern = new QedaPattern(this);
      }
    }

    
    // Check whether number is float

    isFloat(n) {
      return Number(n) && (n % 1 !== 0);
    }

    
    // Merge two objects

    mergeObjects(dest, src) {
      var k, results, v;
      results = [];
      for (k in src) {
        v = src[k];
        if (typeof v === 'object' && (!Array.isArray(v)) && dest.hasOwnProperty(k)) {
          results.push(this.mergeObjects(dest[k], v));
        } else {
          results.push(dest[k] = v);
        }
      }
      return results;
    }

    
    // Parse pin/group list

    parseMultiple(input) {
      var begin, cap, end, group, groups, i, l, len, m, ref, ref1, result;
      if (input == null) {
        return [0];
      }
      result = [];
      groups = input.toString().replace(/\s+/g, '').split(',');
      for (l = 0, len = groups.length; l < len; l++) {
        group = groups[l];
        cap = /(\D*)(\d+)-(\d+)/.exec(group);
        if (cap) {
          begin = parseInt(cap[2]);
          end = parseInt(cap[3]);
          for (i = m = ref = begin, ref1 = end; (ref <= ref1 ? m <= ref1 : m >= ref1); i = ref <= ref1 ? ++m : --m) {
            result.push(cap[1] + i);
          }
        } else {
          result.push(group);
        }
      }
      return result;
    }

    
    // Parse pin numbers

    parsePinNumbers(input) {
      var cap, col, col1, col2, l, len, m, number, numbers, o, ref, ref1, ref2, ref3, result, row, row1, row2;
      if (input == null) {
        return [];
      }
      result = [];
      numbers = input.replace(/\s+/g, '').split(',');
      for (l = 0, len = numbers.length; l < len; l++) {
        number = numbers[l];
        cap = /([A-Z]*)(\d+)-([A-Z]*)(\d+)/.exec(number);
        if (!cap) {
          result.push(number);
        } else {
          row1 = this.gridLetters.indexOf(cap[1]);
          row2 = this.gridLetters.indexOf(cap[3]);
          if (row2 === '') {
            row2 = row1;
          }
          col1 = parseInt(cap[2]);
          col2 = parseInt(cap[4]);
          for (row = m = ref = row1, ref1 = row2; (ref <= ref1 ? m <= ref1 : m >= ref1); row = ref <= ref1 ? ++m : --m) {
            for (col = o = ref2 = col1, ref3 = col2; (ref2 <= ref3 ? o <= ref3 : o >= ref3); col = ref2 <= ref3 ? ++o : --o) {
              result.push(this.gridLetters[row] + col);
            }
          }
        }
      }
      return result;
    }

    
    // Generate symbols and patterns

    render() {
      var bodyPosition, cap, dims, dirName, error, fileName, handler, k, l, len, len1, m, option, options, outline, paths, ref, ref1, ref2, symbol, v;
      if (this._rendered == null) {
        this._rendered = false;
      }
      if (this._rendered) {
        return;
      }
      // Symbols processing
      if (((ref = this.schematic) != null ? ref.symbol : void 0) != null) {
        paths = [`./symbol/${this.library.symbol.style.toLowerCase()}/${this.schematic.symbol.toLowerCase()}`, `./symbol/default/${this.schematic.symbol.toLowerCase()}`, process.cwd() + `/symbol/${this.schematic.symbol.toLowerCase()}`];
        [handler, error] = this._firstHandler(paths);
        ref1 = this.symbols;
        for (l = 0, len = ref1.length; l < len; l++) {
          symbol = ref1[l];
          log.start("Schematic symbol for '" + this.name + (symbol.name != null ? ': ' + symbol.name : '') + "'");
          if (error) {
            log.exception(error);
          }
          handler(symbol, this);
          log.ok();
        }
      }
      // Pattern processing
      if (this.pattern != null) {
        log.start(`Land pattern for '${this.name}'`);
        if (this.housing.options != null) {
          options = this.housing.options.replace(/\s+/g, '').toLowerCase().split(',');
          for (m = 0, len1 = options.length; m < len1; m++) {
            option = options[m];
            this.housing[option] = true;
          }
        }
        this._convertDimensions(this.housing);
        if (((ref2 = this.housing) != null ? ref2.outline : void 0) != null) {
          cap = this.housing.outline.split(' ');
          dirName = cap.shift().toLowerCase();
          fileName = cap.shift().toLowerCase();
          log.start(`Outline '${this.housing.outline}'`);
          try {
            outline = yaml.safeLoad(fs.readFileSync(__dirname + `/../share/outline/${dirName}/${fileName}.yaml`));
          } catch (error1) {
            error = error1;
            log.error(error.message);
          }
          dims = this._processOutline(outline, cap);
          this._convertDimensions(dims);
          for (k in dims) {
            v = dims[k];
            if (this.housing[k] == null) {
              this.housing[k] = v;
            }
          }
          log.ok();
        }
        this._legacy(this.housing); // TODO: Remove in v1.0
        paths = [`./pattern/${this.library.pattern.style.toLowerCase()}/${this.housing.pattern.toLowerCase()}`, `./pattern/default/${this.housing.pattern.toLowerCase()}`, process.cwd() + `/pattern/${this.housing.pattern.toLowerCase()}`];
        [handler, error] = this._firstHandler(paths);
        if (error) {
          log.exception(error);
        }
        handler(this.pattern, this);
        // Box
        this.pattern.box = {
          x: 0,
          y: 0,
          width: this.housing.bodyWidth != null ? this.housing.bodyWidth.nom : 0,
          length: this.housing.bodyLength != null ? this.housing.bodyLength.nom : 0,
          height: this.housing.height != null ? this.housing.height.max : 0
        };
        if (this.housing.bodyPosition != null) {
          bodyPosition = this.pattern.parsePosition(this.housing.bodyPosition);
          this.pattern.box.x = bodyPosition[0].x;
          this.pattern.box.y = bodyPosition[0].y;
        }
        log.ok();
      }
      return this._rendered = true;
    }

    
    // Add pin objects

    _addPins(name, pinOrGroup) {
      var base, i, k, l, len, len1, m, names, number, pinName, pins, result, v;
      result = [];
      if (pinOrGroup == null) {
        return result;
      }
      if (typeof pinOrGroup === 'object') { // Group of pins
        if ((base = this.pinGroups)[name] == null) {
          base[name] = [];
        }
        for (k in pinOrGroup) {
          v = pinOrGroup[k];
          pinName = this.delimiter[name] != null ? name + this.delimiter[name] + k : k;
          pins = this._addPins(pinName, v);
          this.pinGroups[name] = this.pinGroups[name].concat(pins);
          result = result.concat(pins); // Pin number(s)
        }
      } else {
        if (typeof pinOrGroup === 'number') {
          pinOrGroup = pinOrGroup.toString();
        }
        result = result.concat(this.parsePinNumbers(pinOrGroup));
        names = this.parseMultiple(name);
        names.map((v) => {
          var base1;
          return (base1 = this.pinGroups)[v] != null ? base1[v] : base1[v] = [];
        });
        if (names.length > 1) {
          // Dearraying
          if (names.length !== result.length) {
            console.error('Error: Pin count does not correspond pad count');
            process.exit(1);
          }
          for (i = l = 0, len = names.length; l < len; i = ++l) {
            v = names[i];
            this.pins[result[i]] = this._pinObj(result[i], v);
            this.pinGroups[v] = this.pinGroups[v].concat(result[i]);
          }
        } else {
          this.pinGroups[name] = this.pinGroups[name].concat(result);
          for (m = 0, len1 = result.length; m < len1; m++) {
            number = result[m];
            this.pins[number] = this._pinObj(number, name);
          }
        }
      }
      return result;
    }

    
    // Concatenate groups

    _concatenateGroups(groups) {
      var group, l, len, pinGroup, result;
      result = [];
      groups = this.parseMultiple(groups);
      for (l = 0, len = groups.length; l < len; l++) {
        group = groups[l];
        pinGroup = this.pinGroups[group];
        if (pinGroup != null) {
          result = result.concat(pinGroup);
        }
      }
      return result;
    }

    
    // Make dimensions more convenient

    _convertDimensions(housing) {
      var dimensions, k, max, min, nom, tol, v;
      dimensions = ['bodyDiameter', 'bodyLength', 'bodyWidth', 'bodyHeight', 'bossDiameter', 'columnSpan', 'height', 'leadDiameter', 'leadHeight', 'leadLength', 'leadSpan', 'leadWidth', 'leadDiameter1', 'leadHeight1', 'leadLength1', 'leadWidth1', 'leadDiameter2', 'leadHeight2', 'leadLength2', 'leadWidth2', 'pullBack', 'rowSpan', 'tabWidth', 'tabLength'];
      for (k in housing) {
        v = housing[k];
        if (typeof v === 'string' && (/^\d+(\.\d+)?-\d+(\.\d+)?$/.test(v))) {
          v = v.replace(/\s+/g, '').split('-').map(function(v) {
            return parseFloat(v);
          });
        }
        if (Array.isArray(v) && v.length > 0) {
          min = v[0];
          max = v.length > 1 ? v[1] : min;
          nom = (max + min) / 2;
          tol = max - min;
          housing[k] = {
            min: min,
            max: max,
            nom: nom,
            tol: tol
          };
        } else if (dimensions.indexOf(k) !== -1) {
          min = nom = max = v;
          tol = 0;
          housing[k] = {
            min: min,
            max: max,
            nom: nom,
            tol: tol
          };
        }
      }
      if (housing.units === 'inches') {
        this._inchToMm(housing);
        return delete housing.units;
      }
    }

    
    // Return first valid handler

    _firstHandler(paths) {
      var error, handler, handlerError, handlerPath, l, len;
      for (l = 0, len = paths.length; l < len; l++) {
        handlerPath = paths[l];
        handlerError = null;
        try {
          handler = require(handlerPath);
        } catch (error1) {
          error = error1;
          handlerError = error;
          if (error.code === 'MODULE_NOT_FOUND') {
            continue;
          } else {
            break;
          }
        }
        break;
      }
      return [handler, handlerError];
    }

    
    // Convert inches to millimeters

    _inchToMm(value) {
      var k, roundOff, v;
      if (typeof value === 'object') {
        for (k in value) {
          v = value[k];
          value[k] = this._inchToMm(v);
        }
      } else if (this.isFloat(value)) {
        value *= 25.4;
        roundOff = 0.001;
        value = Math.round(value / roundOff) * roundOff;
      } else if (typeof value === 'string') {
        if (/([-0-9.]+\s*,?\s*)+/.test(value)) {
          value = value.replace(/\s+/g, '').split(',').map((v) => {
            return this._inchToMm(parseFloat(v));
          }).reduce((a, v) => {
            return a + ', ' + v;
          });
        }
      }
      return value;
    }

    
    // Process obsolete parameter names

    _legacy(housing) {
      var convertation, convertations, k, results, v;
      convertations = [[/^drillDiameter(.*)/, 'holeDiameter$1']];
      results = [];
      for (k in housing) {
        v = housing[k];
        results.push((function() {
          var l, len, results1;
          results1 = [];
          for (l = 0, len = convertations.length; l < len; l++) {
            convertation = convertations[l];
            if (k.match(convertation[0])) {
              k = k.replace(convertation[0], convertation[1]);
              results1.push(housing[k] = v);
            } else {
              results1.push(void 0);
            }
          }
          return results1;
        })());
      }
      return results;
    }

    
    // Add properties to pins

    _parseProperties() {
      var group, groups, l, len, prop, props, results;
      if (this.properties != null) {
        props = ['analog', 'bidir', 'ground', 'in', 'inverted', 'nc', 'out', 'passive', 'power', 'z'];
        results = [];
        for (l = 0, len = props.length; l < len; l++) {
          prop = props[l];
          if (this.properties[prop] != null) {
            groups = this.parseMultiple(this.properties[prop]);
            results.push((function() {
              var len1, m, ref, results1;
              results1 = [];
              for (m = 0, len1 = groups.length; m < len1; m++) {
                group = groups[m];
                results1.push((ref = this.pinGroups[group]) != null ? ref.map((v) => {
                  var ref1;
                  return (ref1 = this.pins[v]) != null ? ref1[prop] = true : void 0;
                }) : void 0);
              }
              return results1;
            }).call(this));
          } else {
            results.push(void 0);
          }
        }
        return results;
      }
    }

    
    // Generate pin object

    _pinObj(number, name) {
      var obj;
      obj = this.pins[number];
      if (obj != null) {
        if (!this.library.symbol.shortPinNames) {
          obj.name += '/' + name;
        }
      } else {
        obj = {
          name: name,
          number: number
        };
      }
      return obj;
    }

    
    // Add dimensions from outline

    _processOutline(outline, subkeys, result = {}) {
      var k, re, sk, v, valueType;
      if (outline == null) {
        return result;
      }
      sk = subkeys.shift(); // TODO: Create copy of 'subkeys'
      for (k in outline) {
        v = outline[k];
        valueType = (typeof v === 'number') || (typeof v === 'string');
        if (valueType) {
          if (result[k] == null) {
            result[k] = v;
          }
        } else {
          if (sk == null) {
            return;
          }
          re = new RegExp('^' + k + '$');
          if (re.test(sk)) {
            this._processOutline(outline[k], subkeys, result);
          }
        }
      }
      return result;
    }

  };

  module.exports = QedaElement;

}).call(this);
