// Generated by CoffeeScript 2.3.2
(function() {
  var GedaGenerator, KicadGenerator, QedaElement, QedaLibrary, SvgGenerator, fs, log, mkdirp, path, request, yaml;

  fs = require('fs');

  mkdirp = require('mkdirp');

  path = require('path');

  request = require('sync-request');

  yaml = require('js-yaml');

  KicadGenerator = require('./kicad-generator');

  SvgGenerator = require('./svg-generator');

  GedaGenerator = require('./geda-generator');

  QedaElement = require('./qeda-element');

  log = require('./qeda-log');

  QedaLibrary = class QedaLibrary {
    
    // Constructor

    constructor(config = {}) {
      this.output = 'kicad';
      this.connection = {
        timeout: 5000
      };
      this.symbol = {
        style: 'default', // Options: default, GOST
        gridSize: 2.5, // Grid cell size
        factor: 1, // Symbol factor
        pitch: 5, // Default pin pitch
        shortPinNames: false, // Add alternate pin functions to pin name or not
        fill: 'foreground', // Options: none, background, foreground
        fontSize: {
          default: 2.5,
          refDes: 2.5,
          name: 2.5,
          pin: 2.5
        },
        lineWidth: {
          default: 0,
          thick: 0.6,
          thin: 0.2
        },
        space: {
          default: 2,
          pin: 2,
          attribute: 1.5
        }
      };
      this.pattern = {
        style: 'default', 
        densityLevel: 'N', // Nominal
        decimals: 3,
        polarityMark: 'dot',
        preferManufacturer: true,
        smoothPadCorners: false, // Use ratio.cornerToWidth and maximum.cornerRadius
        tolerance: {
          default: 0.1,
          fabrication: 0.1,
          placement: 0.1
        },
        clearance: {
          padToSilk: 0.2,
          padToPad: 0.2,
          padToMask: 0.05,
          leadToHole: 0.1
        },
        ratio: {
          padToHole: 1.5,
          cornerToWidth: 0.25
        },
        minimum: {
          ringWidth: 0.2,
          holeDiameter: 0.2,
          maskWidth: 0.2,
          spaceForIron: 0
        },
        maximum: {
          cornerRadius: 0.2
        },
        lineWidth: {
          default: 0.2,
          silkscreen: 0.12, // IPC-7351C (0.1, 0.12, 0.15)
          assembly: 0.1, // IPC-7351C
          courtyard: 0.05 // IPC-7351C
        },
        fontSize: {
          default: 1,
          refDes: 1.2, // IPC-7351C (1, 1.2, 1.5)
          value: 1
        },
        ball: {
          collapsible: true
        }
      };
      this.mergeObjects(this, config);
      this.pattern.minimum.drillDiameter = this.pattern.minimum.holeDiameter; // TODO: Remove in v1.0
      this.elements = [];
    }

    
    // Add element

    add(element) {
      var alias, aliases, base1, elements, j, l, len, len1, name, newElement, obj, objs, ref, suffixes;
      objs = this.load(element);
      elements = [];
      for (j = 0, len = objs.length; j < len; j++) {
        obj = objs[j];
        if (obj.abstract) {
          console.warn(`'${element}': Cannot add abstract component, use it only as base for others`);
          return elements;
        }
        // Suffixes
        if ((base1 = obj.housing).suffix == null) {
          base1.suffix = '';
        }
        suffixes = obj.housing.suffix.replace(/\s+/g, '').split(',');
        name = obj.name;
        obj.name = name + suffixes[0];
        // Aliases
        aliases = [];
        if (obj.alias != null) {
          ref = obj.alias.replace(/\s+/g, '').split(',');
          for (l = 0, len1 = ref.length; l < len1; l++) {
            alias = ref[l];
            aliases = aliases.concat(suffixes.map((v) => {
              return alias + v;
            }));
          }
        }
        if (suffixes.length > 1) {
          aliases = aliases.concat(suffixes.slice(1).map((v) => {
            return name + v;
          }));
        }
        obj.aliases = aliases;
        newElement = new QedaElement(this, obj);
        this.elements.push(newElement);
        elements.push(newElement);
      }
      return elements;
    }

    
    // Add symbol

    addSymbol(symbol, name, options = "") {
      var def, i, j, newElement, ref, schematic, words;
      schematic = {
        symbol: symbol,
        options: options
      };
      words = name.split('/');
      if (words.length > 1) {
        for (i = j = 0, ref = words.length - 2; (0 <= ref ? j <= ref : j >= ref); i = 0 <= ref ? ++j : --j) {
          schematic[words[i].toLowerCase()] = true;
        }
        name = words[words.length - 1];
      }
      def = {
        name: name,
        schematic: schematic
      };
      newElement = new QedaElement(this, def);
      this.elements.push(newElement);
      return newElement;
    }

    
    // Check element description for validity

    check(obj) {}

    
    // Generate library in given format

    // TODO: Add checking
    generate(name) {
      var generator;
      log.start(`Render library '${name}'`);
      this.render();
      log.ok();
      generator = null;
      switch (this.output) {
        case 'kicad':
          generator = new KicadGenerator(this);
          break;
        case 'svg':
          generator = new SvgGenerator(this);
          break;
        case 'geda':
          generator = new GedaGenerator(this);
      }
      if (generator) {
        log.start(`Generate output for '${name}'`);
        generator.generate(name);
        return log.ok();
      }
    }

    
    // Add ground symbol

    ground(name) {
      return this.addSymbol('ground', name);
    }

    
    // Load element description from remote repository

    load(element, force = false) {
      var j, k, len, modifier, obj, objs, param, v, varObj, variation, variations;
      obj = this.loadYaml(element, force);
      objs = [];
      if (obj.variations != null) {
        variations = obj.variations.replace(/\s+/g, '').toLowerCase().split(',');
        delete obj.variations;
        for (j = 0, len = variations.length; j < len; j++) {
          variation = variations[j];
          if ((typeof filterVariation !== "undefined" && filterVariation !== null) && (variation !== filterVariation)) {
            continue;
          }
          varObj = {};
          this.mergeObjects(varObj, obj);
          for (k in varObj) {
            v = varObj[k];
            [param, modifier] = k.replace(/\s+/g, '').split('@');
            if (modifier != null) {
              if (modifier.toLowerCase() === variation) {
                varObj[param] = v; // Replace
              }
              delete varObj[k];
            }
          }
          objs.push(varObj);
        }
      } else {
        objs.push(obj);
      }
      return objs;
    }

    
    // Load and parse YAML file

    loadYaml(element, force = false) {
      var base, baseElement, baseObj, bases, elementName, elementYaml, error, exclusion, exclusions, filterVariation, j, l, len, len1, localFile, obj, res;
      element = element.toLowerCase();
      [elementName, filterVariation] = element.split('@');
      elementYaml = elementName + '.yaml';
      localFile = './library/' + elementYaml;
      if ((!fs.existsSync(localFile)) || force) {
        log.start(`Load '${elementName}'`);
        elementYaml = elementYaml.split('/').map(function(v) {
          return encodeURIComponent(v);
        }).join('/');
        try {
          res = request('GET', 'https://raw.githubusercontent.com/qeda/library/master/' + elementYaml, {
            timeout: this.connection.timeout
          });
        } catch (error1) {
          error = error1;
          log.error(error.message);
        }
        if (res.statusCode === 200) {
          mkdirp.sync(path.dirname(localFile));
          fs.writeFileSync(localFile, res.body);
          log.ok();
        } else {
          log.warning(res.statusCode);
        }
      }
      log.start(`Read '${elementName}'`);
      try {
        obj = yaml.safeLoad(fs.readFileSync(localFile));
      } catch (error1) {
        // TODO: YAML Schema validation
        error = error1;
        log.error(`${error.message}`);
      }
      log.ok();
      this.check(obj);
      // Load base descrition
      if (obj.base != null) {
        if (typeof obj.base !== 'string') {
          obj.base = obj.base.toString();
        }
        bases = obj.base.replace(/\s+/g, '').split(',');
        delete obj.base; // We do not need this information now
        exclusions = [
          'abstract',
          'alias' // Exclude these fields from base object
        ];
        for (j = 0, len = bases.length; j < len; j++) {
          base = bases[j];
          baseElement = base;
          if (path.dirname(baseElement) === '.') {
            baseElement = path.dirname(element) + '/' + baseElement;
          }
          baseObj = this.loadYaml(baseElement, force);
          for (l = 0, len1 = exclusions.length; l < len1; l++) {
            exclusion = exclusions[l];
            if (baseObj[exclusion] != null) {
              delete baseObj[exclusion];
            }
          }
          this.mergeObjects(baseObj, obj);
          obj = baseObj;
        }
      }
      if (element.indexOf('/') !== -1) {
        if (obj.group == null) {
          obj.group = element.substr(0, element.indexOf('/')).toUpperCase();
        }
      }
      return obj;
    }

    
    // Merge two objects

    mergeObjects(dest, src) {
      var k, results, v;
      results = [];
      for (k in src) {
        v = src[k];
        if (typeof v === 'object' && (!Array.isArray(v)) && dest.hasOwnProperty(k)) {
          results.push(this.mergeObjects(dest[k], v));
        } else {
          results.push(dest[k] = v);
        }
      }
      return results;
    }

    
    // Add power symbol

    power(name) {
      return this.addSymbol('power', name);
    }

    
    // Add port symbols

    port() {
      var input, output;
      input = this.addSymbol('port', 'inputExt', 'input');
      return output = this.addSymbol('port', 'outputExt', 'output');
    }

    
    // Render elements

    render() {
      var baseName, element, i, j, keys, len, name, patterns, ref;
      if (this._rendered == null) {
        this._rendered = false;
      }
      if (this._rendered) {
        return;
      }
      patterns = {};
      ref = this.elements;
      for (j = 0, len = ref.length; j < len; j++) {
        element = ref[j];
        element.render();
        // Check for pattern names duplication
        if (element.pattern != null) {
          baseName = element.pattern.name;
          name = baseName;
          keys = Object.keys(patterns);
          i = 1;
          while (keys.indexOf(name) !== -1) {
            if (element.pattern.isEqualTo(patterns[name])) {
              break;
            } else {
              name = `${baseName}-${i++}`;
              element.pattern.name = name;
            }
          }
          patterns[name] = element.pattern;
        }
      }
      return this._rendered = true;
    }

  };

  module.exports = QedaLibrary;

}).call(this);
