// Generated by CoffeeScript 2.3.2
(function() {
  var SvgGenerator, boundingBox, describeArc, describePoly, fs, iterPoints, log, mkdirp, netlistsvg_builtins, netlistsvg_properties, netlistsvg_style, pcbsvg_style, polarToCartesian, rotateShape, rotateSymbol, sprintf, svg_head, svg_tail, translate;

  fs = require('fs');

  mkdirp = require('mkdirp');

  sprintf = require('sprintf-js').sprintf;

  log = require('./qeda-log');

  svg_head = "<svg xmlns=\"http://www.w3.org/2000/svg\"\n     xmlns:xlink=\"http://www.w3.org/1999/xlink\"\n     xmlns:s=\"https://github.com/nturley/netlistsvg\"";

  svg_tail = "</svg>\n";

  netlistsvg_properties = "<s:properties\n  constants=\"false\"\n  splitsAndJoins=\"false\"\n  genericsLaterals=\"true\">\n  <s:layoutEngine\n    org.eclipse.elk.layered.spacing.nodeNodeBetweenLayers=\"5\"\n    elk.spacing.nodeNode= \"35\"\n    org.eclipse.elk.direction=\"DOWN\"/>\n  <s:low_priority_alias val=\"$dff\" />\n</s:properties>";

  netlistsvg_style = "<style>\nsvg {\n  stroke: #000;\n  fill: none;\n  stroke-linejoin: round;\n  stroke-linecap: round;\n}\ntext {\n  fill: #000;\n  stroke: none;\n  font-family: \"Courier New\", monospace;\n}\n</style>\n";

  pcbsvg_style = "<style>\nsvg {\n  stroke: #000;\n  fill: none;\n}\ntext {\n  fill: #000;\n  stroke: none;\n  font-family: \"Courier New\", monospace;\n}\n.topCopper {\n  fill: red;\n}\n.bottomCopper {\n  fill: green;\n}\n.topSilkscreen, .bottomSilkscreen {\n  stroke: yellow;\n  stroke-linejoin: round;\n  stroke-linecap: round;\n}\n.topCourtyard, .bottomCourtyard {\n  stroke: blue;\n  stroke-linejoin: round;\n  stroke-linecap: round;\n}\n.topAssembly, .bottomAssembly {\n  stroke: purple;\n  stroke-linejoin: round;\n  stroke-linecap: round;\n}\n</style>\n";

  netlistsvg_builtins = "<!-- builtin -->\n<g s:type=\"generic\" s:width=\"30\" s:height=\"40\" transform=\"translate(40,20)\">\n  <s:alias val=\"ic\"/>\n  <text x=\"15\" y=\"-4\" s:attribute=\"ref\" text-anchor=\"middle\" font-size=\"10\">generic</text>\n  <rect width=\"30\" height=\"40\" x=\"0\" y=\"0\" s:generic=\"body\"/>\n  <g transform=\"translate(30,10)\"\n     s:x=\"30\" s:y=\"10\" s:pid=\"out0\" s:position=\"left\">\n    <text x=\"5\" y=\"-4\" font-size=\"10\" text-anchor=\"start\">out0</text>\n  </g>\n  <g transform=\"translate(30,30)\"\n     s:x=\"30\" s:y=\"30\" s:pid=\"out1\" s:position=\"left\">\n    <text x=\"5\" y=\"-4\" font-size=\"10\" text-anchor=\"start\">out1</text>\n  </g>\n  <g transform=\"translate(0,10)\"\n     s:x=\"0\" s:y=\"10\" s:pid=\"in0\" s:position=\"right\">\n      <text x=\"-3\" y=\"-4\" font-size=\"10\" text-anchor=\"end\">in0</text>\n  </g>\n  <g transform=\"translate(0,30)\"\n     s:x=\"0\" s:y=\"30\" s:pid=\"in1\" s:position=\"right\">\n    <text x=\"-3\" y=\"-4\" font-size=\"10\" text-anchor=\"end\">in1</text>\n  </g>\n</g>\n<!-- builtin -->\n";

  polarToCartesian = function(centerX, centerY, radius, angleInDegrees) {
    var angleInRadians;
    angleInRadians = (angleInDegrees - 90) * Math.PI / 180.0;
    return {
      x: centerX + (radius * Math.cos(angleInRadians)),
      y: centerY + (radius * Math.sin(angleInRadians))
    };
  };

  describeArc = function(shape) {
    var a, end, largeArcFlag, m, start;
    start = polarToCartesian(shape.x, shape.y, shape.radius, shape.end + 90);
    end = polarToCartesian(shape.x, shape.y, shape.radius, shape.start + 90);
    largeArcFlag = shape.end - shape.start <= 180 ? "0" : "1";
    m = `M${start.x},${start.y}`;
    a = `A${shape.radius},${shape.radius} 0 ${largeArcFlag} 0 ${end.x},${end.y}`;
    return `${m} ${a}`;
  };

  describePoly = function(shape) {
    var cmd, d, ref, x, y, z;
    d = [];
    ref = iterPoints(shape);
    for (z of ref) {
      [x, y] = z;
      cmd = d.length === 0 ? 'M' : 'L';
      d.push(`${cmd}${x},${y}`);
    }
    return d.join(' ');
  };

  boundingBox = function(iterator) {
    var maxx, maxy, minx, miny, ref, x, y, z;
    minx = maxx = miny = maxy = 0;
    ref = iterator();
    for (z of ref) {
      [x, y] = z;
      minx = Math.min(minx, x);
      maxx = Math.max(maxx, x);
      miny = Math.min(miny, y);
      maxy = Math.max(maxy, y);
    }
    return [minx, miny, maxx, maxy];
  };

  iterPoints = function*(shape) {
    var i, ref, ref1, results;
    switch (shape.kind) {
      case 'pin':
        return (yield [shape.x, shape.y]);
      case 'pad':
        yield [shape.x + shape.width / 2, shape.y + shape.height / 2];
        return (yield [shape.x - shape.width / 2, shape.y - shape.height / 2]);
      case 'attribute':
        yield [shape.x, shape.y + shape.fontSize];
        return (yield [shape.x, shape.y - shape.fontSize]);
      case 'line':
      case 'rectangle':
        yield [shape.x1, shape.y1];
        return (yield [shape.x2, shape.y2]);
      case 'circle':
      case 'arc':
        yield [shape.x - shape.radius, shape.y - shape.radius];
        return (yield [shape.x + shape.radius, shape.y + shape.radius]);
      case 'poly':
        ref1 = (function() {
          var results1 = [];
          for (var j = 0, ref = shape.points.length - 1; 0 <= ref ? j <= ref : j >= ref; 0 <= ref ? j++ : j--){ results1.push(j); }
          return results1;
        }).apply(this);
        results = [];
        for (i of ref1) {
          if (i % 2 === 0) {
            results.push((yield [shape.points[i], shape.points[i + 1]]));
          } else {
            results.push(void 0);
          }
        }
        return results;
    }
  };

  translate = function(shift_x, shift_y) {
    return function(coord) {
      return [coord[0] + shift_x, coord[1] + shift_y];
    };
  };

  rotateShape = function(shape, angle, transformer) {
    var dx, dxp, dy, dyp, i, ref, ref1, results, rotate, rotator, transform;
    rotator = {
      90: function(c) {
        return [-c[1], c[0]];
      },
      180: function(c) {
        return [-c[0], -c[1]];
      },
      270: function(c) {
        return [c[1], -c[0]];
      }
    };
    rotate = rotator[angle];
    switch (shape.kind) {
      case 'pin':
        [shape.x, shape.y] = transformer([shape.x, shape.y], rotate);
        [shape.xdir, shape.ydir] = rotate([shape.xdir, shape.ydir]);
        if (shape.xdir === 0) {
          if (shape.ydir === 1) {
            shape.orientation = 'down';
            return shape.position = 'top';
          } else {
            shape.orientation = 'up';
            return shape.position = 'bottom';
          }
        } else {
          if (shape.xdir === 1) {
            shape.orientation = 'left';
            return shape.position = 'right';
          } else {
            shape.orientation = 'right';
            return shape.position = 'left';
          }
        }
        break;
      case 'line':
        [shape.x1, shape.y1] = transformer([shape.x1, shape.y1], rotate);
        return [shape.x2, shape.y2] = transformer([shape.x2, shape.y2], rotate);
      case 'circle':
        return [shape.x, shape.y] = transformer([shape.x, shape.y], rotate);
      case 'attribute':
        [shape.x, shape.y] = transformer([shape.x, shape.y], rotate);
        [shape.dx, shape.dy] = rotate([shape.dx, shape.dy]);
        if (angle === '90' || angle === '270') {
          shape.writing_mode = 'tb';
        }
        if (angle === '180' || angle === '270') {
          if (shape.anchor === 'start') {
            return shape.anchor = 'end';
          } else if (shape.anchor === 'end') {
            return shape.anchor = 'start';
          }
        }
        break;
      case 'rectangle':
        dx = dxp = (shape.x2 - shape.x1) / 2;
        dy = dyp = (shape.y2 - shape.y1) / 2;
        if (angle === '90' || angle === '270') {
          [dxp, dyp] = [dyp, dxp];
        }
        transform = function(c) {
          return translate(-dxp, -dyp)(rotate(translate(dx, dy)(c)));
        };
        [shape.x1, shape.y1] = transformer([shape.x1, shape.y1], transform);
        return [shape.x2, shape.y2] = transformer([shape.x2, shape.y2], transform);
      case 'arc':
        [shape.x, shape.y] = transformer([shape.x, shape.y], rotate);
        shape.start += Number(angle);
        return shape.end += Number(angle);
      case 'poly':
        ref1 = (function() {
          var results1 = [];
          for (var j = 0, ref = shape.points.length - 1; 0 <= ref ? j <= ref : j >= ref; 0 <= ref ? j++ : j--){ results1.push(j); }
          return results1;
        }).apply(this);
        results = [];
        for (i of ref1) {
          if (i % 2 === 0) {
            results.push([shape.points[i], shape.points[i + 1]] = transformer([shape.points[i], shape.points[i + 1]], rotate));
          } else {
            results.push(void 0);
          }
        }
        return results;
    }
  };

  rotateSymbol = function(symbol, angle) {
    var cx, cxp, cy, cyp, transformer;
    if (typeof angle !== 'string') {
      angle = String(angle);
    }
    if (angle === '0') {
      return;
    }
    cx = cxp = symbol.width / 2;
    cy = cyp = symbol.height / 2;
    if (angle === '90' || angle === '270') {
      [cxp, cyp] = [cyp, cxp];
      [symbol.width, symbol.height] = [symbol.height, symbol.width];
    }
    transformer = function(c, transform) {
      return translate(cxp, cyp)(transform(translate(-cx, -cy)(c)));
    };
    return symbol.shapes.forEach(function(shape) {
      return rotateShape(shape, angle, transformer);
    });
  };

  
  // Generator of library in SVG format

  SvgGenerator = class SvgGenerator {
    
    // Constructor

    constructor(library) {
      this.library = library;
      this.dir = './svg';
      this.f = `%.${this.library.pattern.decimals}f`;
      this.x_pos = 120;
      this.y_pos = 20;
      this.max_height = 40;
      this.spacing = 20;
    }

    
    // Generate symbol library and footprint files

    generate(name1) {
      var aliases, element, fd, j, k, l, len, len1, len2, name, options, orientation, pattern, patternName, patterns, ref, ref1, ref2, ref3, results, showPinNames, showPinNumbers, string, sym, symbol, symbol_lookup, symbols;
      this.name = name1;
      symbols = [];
      symbol_lookup = {};
      patterns = {};
      ref = this.library.elements;
      for (j = 0, len = ref.length; j < len; j++) {
        element = ref[j];
        symbol = element.schematic.symbol;
        options = element.schematic.options;
        string = `${symbol} ${(options != null ? options : '')}`;
        name = (string.match(/[^\s,]+/g) || []).join('-').toLowerCase();
        showPinNames = ((ref1 = element.schematic) != null ? ref1.showPinNames : void 0) ? true : false;
        showPinNumbers = ((ref2 = element.schematic) != null ? ref2.showPinNumbers : void 0) ? true : false;
        if (name.startsWith('connector') || name.startsWith('ic')) {
          // Use generic component
          continue;
        }
        aliases = [element.name].concat(element.aliases);
        if (!(showPinNames || showPinNumbers)) {
          aliases = [name].concat(aliases);
        }
        if (!(aliases[0] in symbol_lookup)) {
          symbol_lookup[aliases[0]] = symbols.length;
          symbols.push({
            'name': name,
            'aliases': aliases,
            // doesn't handle multipart symbols yet
            'symbol': element.symbols[0],
            'showPinNames': showPinNames,
            'showPinNumbers': showPinNumbers
          });
        } else {
          sym = symbols[symbol_lookup[aliases[0]]];
          sym.aliases = sym.aliases.concat(aliases);
        }
        if (element.pattern != null) {
          patterns[element.pattern.name] = element.pattern;
        }
      }
      // Symbols
      log.start(`SVG symbols skin '${this.name}.svg'`);
      mkdirp.sync(`${this.dir}`);
      fd = fs.openSync(`${this.dir}/${this.name}.svg`, 'w');
      fs.writeSync(fd, svg_head + '>\n');
      fs.writeSync(fd, netlistsvg_properties);
      fs.writeSync(fd, netlistsvg_style);
      fs.writeSync(fd, netlistsvg_builtins);
      for (k = 0, len1 = symbols.length; k < len1; k++) {
        symbol = symbols[k];
        ref3 = symbol.symbol.orientations;
        for (l = 0, len2 = ref3.length; l < len2; l++) {
          orientation = ref3[l];
          this._generateSymbol(fd, symbol.name, symbol.aliases, symbol.symbol, orientation, symbol.showPinNames, symbol.showPinNumbers);
        }
      }
      fs.writeSync(fd, svg_tail);
      fs.closeSync(fd);
      log.ok();
// Footprints
      results = [];
      for (patternName in patterns) {
        pattern = patterns[patternName];
        log.start(`SVG footprint '${patternName}.svg'`);
        mkdirp.sync(`${this.dir}/${this.name}`);
        fd = fs.openSync(`${this.dir}/${this.name}/${patternName}.svg`, 'w');
        this._generatePattern(fd, pattern);
        fs.writeSync(fd, svg_tail);
        fs.closeSync(fd);
        results.push(log.ok());
      }
      return results;
    }

    
    // Write symbol entry to svg file

    _generateSymbol(fd, name, aliases, symbol, orientation, showPinNames, showPinNumbers) {
      var clone, height, j, len, minx, miny, orientation_postfix, ref, shape, width;
      clone = JSON.parse(JSON.stringify(symbol));
      ref = clone.shapes;
      // Preprocess symbol
      for (j = 0, len = ref.length; j < len; j++) {
        shape = ref[j];
        this._preprocessShape(shape, this.library.symbol, false);
      }
      // Bounding box
      [minx, miny, width, height] = this._getBoundingBox(clone);
      // Scale and shift and rotate symbol
      clone.width = width;
      clone.height = height;
      symbol.resize.call(clone, 10, true, -minx, -miny);
      rotateSymbol(clone, orientation);
      orientation_postfix = orientation === 0 ? '' : '_rot' + orientation;
      // Symbol start
      fs.writeSync(fd, `<g s:type="${name}${orientation_postfix}"\n   s:width="${clone.width}"\n   s:height="${clone.height}"\n   transform="translate(${this.x_pos}, ${this.y_pos})">\n`);
      // Update skin position for next symbol
      this.x_pos += clone.width + this.spacing;
      this.max_height = Math.max(this.max_height, clone.height);
      if (this.x_pos >= 1200) {
        this.x_pos = this.spacing;
        this.y_pos += this.max_height + this.spacing;
        this.max_height = 0;
      }
      // Write aliases
      aliases.forEach((alias) => {
        var string;
        string = `<s:alias val="${alias}${orientation_postfix}"/>\n`;
        return fs.writeSync(fd, string);
      });
      // Write shapes
      clone.shapes.forEach((shape) => {
        return fs.writeSync(fd, this._shapeToSvg(shape, showPinNames, showPinNumbers));
      });
      // Symbol end
      return fs.writeSync(fd, "</g>\n");
    }

    
    // Write pattern to svg file

    _generatePattern(fd, pattern) {
      var height, j, len, minx, miny, ref, shape, width;
      ref = pattern.shapes;
      // Preprocess pattern
      for (j = 0, len = ref.length; j < len; j++) {
        shape = ref[j];
        this._preprocessShape(shape, this.library.pattern, true);
      }
      // Bounding box
      [minx, miny, width, height] = this._getBoundingBox(pattern);
      fs.writeSync(fd, `${svg_head}\n     width="${width}mm"\n     height="${height}mm"\n     viewBox="0 0 ${width} ${height}">\n`);
      fs.writeSync(fd, pcbsvg_style);
      // Pattern start
      fs.writeSync(fd, `<g transform="translate(${width / 2}, ${height / 2})">\n`);
      // Write shapes
      pattern.shapes.forEach((shape) => {
        return fs.writeSync(fd, this._shapeToSvg(shape));
      });
      // Pattern end
      return fs.writeSync(fd, "</g>\n");
    }

    
    // Returns an SVG representation of a shape

    _shapeToSvg(shape, showPinNames, showPinNumbers) {
      var attributes, dx, dy, pin_text, ref, ref1, ref2, text_anchor, writing_mode;
      if (shape.visible === false) {
        return "";
      }
      switch (shape.kind) {
        case 'pin':
          pin_text = '';
          writing_mode = 'lr';
          text_anchor = 'start';
          if (showPinNames) {
            pin_text = shape.name;
            if ((ref = shape.position) === 'top' || ref === 'bottom') {
              writing_mode = 'tb';
              dx = 0.4;
            }
            if ((ref1 = shape.position) === 'right' || ref1 === 'bottom') {
              text_anchor = 'end';
            }
          } else if (showPinNumbers) {
            pin_text = shape.number;
          }
          dx = 0;
          dy = 0;
          if ((ref2 = shape.position) === 'left' || ref2 === 'right') {
            dy = -0.4;
          } else {
            dx = 0.6;
          }
          return `<g s:x="${shape.x}"\n   s:y="${shape.y}"\n   s:pid="${shape.name}"\n   s:position="${shape.position}">\n  <line x1="${shape.x}"\n        y1="${shape.y}"\n        x2="${shape.x + shape.xdir * shape.length}"\n        y2="${shape.y + shape.ydir * shape.length}"\n        stroke-width="${shape.lineWidth}"/>\n  <text x="${shape.x}"\n        y="${shape.y}"\n        writing-mode="${writing_mode}"\n        text-anchor="${text_anchor}"\n        font-size="${shape.fontSize}"\n        dx="${dx}em"\n        dy="${dy}em">${pin_text}</text>\n</g>\n`;
        case 'pad':
          attributes = `s:x="${shape.x}" s:y="${shape.y}" s:pid="${shape.name}"\nclass="${shape.layer}"`;
          switch (shape.shape) {
            case 'rectangle':
              return `<rect x="${shape.x - shape.width / 2}" y="${shape.y - shape.height / 2}"\n      width="${shape.width}" height="${shape.height}"\n      stroke-width="0"\n      ${attributes}/>\n`;
            case 'circle':
              return `<ellipse cx="${shape.x}" cy="${shape.y}"\n         rx="${shape.width / 2}" ry="${shape.height / 2}"\n         stroke-width="${shape.lineWidth}"\n         ${attributes}/>`;
          }
          break;
        case 'attribute':
          return `<text x="${shape.x}" y="${shape.y}" font-size="${shape.fontSize}"\n      text-anchor="${shape.anchor}" dx="${shape.dx}em" dy="${shape.dy}em"\n      writing-mode="${shape.writing_mode}"\n      s:attribute="${shape.name}"\n      class="${shape.layer || ''}">${shape.name}</text>\n`;
        case 'line':
          return `<line x1="${shape.x1}" x2="${shape.x2}"\n      y1="${shape.y1}" y2="${shape.y2}"\n      stroke-width="${shape.lineWidth}"\n      class="${shape.layer || ''}"/>\n`;
        case 'rectangle':
          return `<rect x="${shape.x1}" y="${shape.y1}"\n      width="${Math.abs(shape.x2 - shape.x1)}"\n      height="${Math.abs(shape.y2 - shape.y1)}"\n      stroke-width="${shape.lineWidth}"\n      class="${shape.layer || ''}"/>\n`;
        case 'circle':
          return `<circle cx="${shape.x}" cy="${shape.y}"\n        r="${shape.radius}" fill="${shape.fill}"\n        stroke-width="${shape.lineWidth}"\n        class="${shape.layer || ''}"/>\n`;
        case 'arc':
          return `<path d="${describeArc(shape)}" stroke-width="${shape.lineWidth}"\n      class="${shape.layer || ''}"/>\n`;
        case 'poly':
          return `<path d="${describePoly(shape)}" fill="${shape.fill}"\n      stroke-width="${shape.lineWidth}"\n      class="${shape.layer || ''}"/>\n`;
      }
    }

    _preprocessShape(shape, settings, hideAttributes) {
      if (shape.lineWidth != null) {
        if (shape.lineWidth === 0) {
          shape.lineWidth = settings.lineWidth.default;
        }
      }
      if (shape.layer != null) {
        shape.layer = shape.layer.join(' ');
      }
      if (shape.fill != null) {
        shape.fill = shape.fill === 'background' ? '#000' : 'none';
      }
      if (shape.valign != null) {
        switch (shape.valign) {
          case "bottom":
            shape.dy = 0;
            break;
          case "center":
            shape.dy = .4;
            break;
          case "top":
            shape.dy = 1;
        }
      }
      if (shape.halign != null) {
        switch (shape.halign) {
          case "left":
            shape.anchor = "start";
            break;
          case "right":
            shape.anchor = "end";
            break;
          case "center":
            shape.anchor = "middle";
        }
      }
      if (shape.kind === 'attribute') {
        shape.writing_mode = 'lr';
        shape.dx = 0;
      }
      if (shape.kind === 'pin') {
        shape.xdir = 0;
        shape.ydir = 0;
        switch (shape.orientation) {
          case 'up':
            shape.position = 'bottom';
            shape.ydir = -1;
            break;
          case 'down':
            shape.position = 'top';
            shape.ydir = 1;
            break;
          case 'right':
            shape.position = 'left';
            shape.xdir = 1;
            break;
          case 'left':
            shape.position = 'right';
            shape.xdir = -1;
        }
      }
      if (shape.kind === 'attribute') {
        if (hideAttributes) {
          shape.visible = false;
        }
        switch (shape.name) {
          case 'refDes':
            if (shape.fontSize == null) {
              shape.fontSize = settings.fontSize.refDes;
            }
            return shape.name = 'ref';
          case 'value':
            return shape.fontSize != null ? shape.fontSize : shape.fontSize = settings.fontSize.value;
          default:
            return shape.fontSize != null ? shape.fontSize : shape.fontSize = settings.fontSize.default;
        }
      }
    }

    _getBoundingBox(hasShapes) {
      var height, maxx, maxy, minx, miny, width;
      [minx, miny, maxx, maxy] = boundingBox(function*() {
        var j, len, ref, results, shape;
        ref = hasShapes.shapes;
        results = [];
        for (j = 0, len = ref.length; j < len; j++) {
          shape = ref[j];
          results.push((yield* iterPoints(shape)));
        }
        return results;
      });
      width = maxx - minx;
      height = maxy - miny;
      return [minx, miny, width, height];
    }

  };

  module.exports = SvgGenerator;

}).call(this);
